#!/usr/bin/perl -I/usr/local/lib/bs/bin
#    Big Sister network monitor
#    Copyright (C) 1998  Thomas Aeby
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

#=============================================================================
#
$Usage	  = "[-D level] [-r rule_file] [-t loop_time] [-s state_dir] [-l logfile] [-b basedir]";
#
#=============================================================================

# Trim the program path name
$prog = $progdir = $0;
$prog =~ s#.*/##;
$progdir =~ s#/[^\/]*$##;
 
use Getopt::Long;
use FileHandle;

$loop_delay = 15*60;
$log_delay = 30;

$Getopt::Long::ignorecase = 0;  
GetOptions( "D:i", "r:s", "t:i", "s:s\@", "b:s", "l:s@" ) || &usage;
$dl = $opt_D;

if( $opt_b ) {
    $root = $opt_b;
}
elsif( -d "/l/lib/bs" ) {
    $root = "/l/lib/bs";
}
elsif( -d "/usr/local/lib/bs" ) {
    $root = "/usr/local/lib/bs";
}
else {
    $root = $progdir."/..";
}

unshift( @INC, $root."/bin" );
eval "use bb_evgen";

&bb_evgen::init;

$ENV{PATH} = "$root/bin:".$ENV{PATH};

@state_dirs = ("$root/www/logs");
$rule_file = "$root/adm/bb_event_generator.cfg";
$bb_evgen::pager = "$root/bin/log_mail";
$bb_evgen::page_user = "alarm";

($rule_file=$opt_r) if( $opt_r );
(@state_dirs=@opt_s) if( @opt_s );
@log_files = @opt_f;
unless( @log_files ) {
    if( -d "$root/www/logs/history" ) {
	@log_files = "$root/www/logs/history/display.history";
    }
}
($loop_delay=5*60) unless( @log_files );
($loop_delay=$opt_t) if( $opt_t );

$bb_evgen::rule_file = $rule_file;
@bb_evgen::state_dirs = @state_dirs;

&bb_evgen::load_rules || exit 1;
&bb_evgen::read_state;
&bb_evgen::open_logs( @log_files );

&background unless( $dl );

$next_poll = time+$loop_delay;

while( 1 ) {
    &bb_evgen::check_rulefile;
    &bb_evgen::check_logs;
    &bb_evgen::check_alarms;
    if( time>=$next_poll ) {
	&bb_evgen::check_state;
	$next_poll += $loop_delay;
    }
    $sleep = $next_poll-time;
    $sleep = $log_delay if( $sleep>$log_delay );
    sleep $sleep if( $sleep>0 );
}

sub usage {
    print "Usage: $prog $Usage\n";
    exit 1;
}


sub background {
    chdir "/";
    close STDIN; close STDOUT; close STDERR;
    if (open(DEVTTY, "/dev/tty")) {
	ioctl(DEVTTY,0x20007471,0);         # XXX s/b &TIOCNOTTY
	close DEVTTY;
    }
    open(STDIN,"<&TTY");
    open(STDOUT,">&TTY");
    open(STDERR,">&STDOUT");
    fork && exit;
    $SIG{"HUP"} = "IGNORE";
    setpriority( "PRIO_PROCESS", 0, 10 );
}
	
