package bb_evgen;

#=============================================================================
#
# Id	  : @(#) $Id: bb_evgen.pm,v 1.1.1.1 1998/11/08 14:02:37 aeby Exp $
# Package : bb
# File	  : bb_evgen
# Author  : Thomas Aeby
# Date	  : 22.10.97
#
#
# $Log: bb_evgen.pm,v $
# Revision 1.1.1.1  1998/11/08 14:02:37  aeby
# release 0.31 imported into CVS
#
#
#=============================================================================

# check: condition met -> do not clear delayed alarm
# delay: delay in minutes
# repeat: warn every x minutes until alarm is cleared ("reminder")
# norepeat: do not raise another alarm for the same host/check within x minutes
# keep: keep alarm for at least x minutes even if it is cleared
# prio: alarm priority
# repeatprio: reminder priority
# mail: comma separated list of mail recipients

use strict;
use FileHandle;
use Socket;

sub init {
    @bb_evgen::state_dirs = ();
    @bb_evgen::logs = ();
    $bb_evgen::pager = "";
    $bb_evgen::pager_user = "";
    $bb_evgen::rule_file = "";
    $bb_evgen::rule_file_stat = "";

    %bb_evgen::numstates = ( "red", 0, "yellow", 1, "purple", 2, "green", 3, "never", -1 );

    @bb_evgen::new_alarms = ();
    @bb_evgen::old_alarms = ();
    @bb_evgen::alarms = ();
    %bb_evgen::up = ();

    $bb_evgen::minute = 10;
}



sub check_rulefile {
    print "check_rulefile()\n" if( $main::dl );
    &bb_evgen::load_rules if( (stat($bb_evgen::rule_file))[9] != $bb_evgen::rule_file_stat );
}


sub check_state {
    print "check_state()\n" if( $main::dl );
    &bb_evgen::read_state;
}


sub check_alarms {
    my( $i, $var, $now );
    print "check_alarms()\n" if( $main::dl );
    $now = time;
    for( $i=$#bb_evgen::new_alarms; $i>=0; $i-- ) {
	$var = $bb_evgen::new_alarms[$i];
	if( ($$var{"delay"}<$now) ) {
	    splice( @bb_evgen::new_alarms, $i, 1 );
	    &bb_evgen::alarm( $var );
	}
    }
    for( $i=$#bb_evgen::alarms; $i>=0; $i-- ) {
	$var = $bb_evgen::alarms[$i];
	if( $$var{"_next_repetition"} && ($$var{"_next_repetition"} < $now) ) {
	    &bb_evgen::alarm( $var, 1 );
	}
	if( $$var{"_dequeue"} && ($$var{"_dequeue"}<$now) 
	    && ($bb_evgen::up{$$var{"host"}.".".$$var{"item"}}) ) {
	    &bb_evgen::clear_alarm( $i );
	}
    }
    for( $i=$#bb_evgen::old_alarms; $i>=0; $i-- ) {
	$var = $bb_evgen::old_alarms[$i];
	if( $$var{"norepeat"} <= $now ) {
	    print "\x07forgetting old alarm: ".$$var{"host"}.".".$$var{"item"}." ".$$var{"text"}."\n" if( $main::dl );
	    splice( @bb_evgen::old_alarms, $i, 1 );
	}
    }
}



sub check_logs {
    my( $log );
    my( $log_file, $fh, $pos, $open, $ino, $buff ); 
    my( $newbuff, $line, $prev, $newnum );

    print "check_logs()\n" if( $main::dl );
    foreach $log (@bb_evgen::logs) {
	( $log_file, $fh, $pos, $open, $ino, $buff ) = @$log; 
	print "checking $log_file\n" if( $main::dl>3 );
	if( $open ) {
	    seek( $fh, $pos, 0 );
	    read( $fh, $newbuff, 200000 );
	    print "read at pos $pos: $newbuff\n" if( $main::dl>7 );
	    $buff .= $newbuff;
	    $pos = tell($fh);
	    if( $ino != ((stat($log_file))[1]) ) {
		print "log file $log_file was replaced\n" if( $main::dl>4 );
		close( $fh );
		$open = 0;
	    }
	}
	unless( $open ) {
	    print "(re-)opening log file $log_file\n" if( $main::dl>4 );
	    if( open( $fh, "<$log_file" ) ) {
		$open = 1;
		$ino = ((stat($log_file))[1]);
		read( $fh, $newbuff, 200000 );
		$buff .= $newbuff;
		$pos = tell($fh);
	    }
	}
	while( $buff =~ /^([^\n]*)\n/ ) {
	    $buff = $';
	    $line = $1;
	    print "line: $line\n" if( $main::dl>5 );
	    if( $line =~ /^([^\.]+)\.([^:]+): ([a-z]+)->([a-z]+)(.*)$/ ) {
		my( $host, $item, $old, $new, $text ) = ($1,$2,$3,$4,$5);
		$newnum = $bb_evgen::numstates{$new};
		$prev = $bb_evgen::previous_state{$host.".".$item};
		print "$host.$item $prev->$newnum($new)\n" if( $main::dl>5 );
		if( $newnum != $prev ) {
		    ($prev = $newnum) unless( defined $prev );
		    &bb_evgen::raise( $host, $item, $bb_evgen::numstates{$old}, 
			$newnum, $new." ".$text );
		}
		$bb_evgen::previous_state{$host.".".$item} = $newnum;
	    }
	}
	@$log = ( $log_file, $fh, $pos, $open, $ino, $buff ); 
    }
}


sub read_state {
    my( $dir, $file, $state, $prev, $color, $fulltext );

    print "read_state()\n" if( $main::dl>1 );
    %bb_evgen::goes_up = %bb_evgen::goes_down = %bb_evgen::is_up = %bb_evgen::is_down = ();
    foreach $dir( @bb_evgen::state_dirs ) {
	print "reading state dir $dir\n" if( $main::dl>2 );
	opendir( DIR, $dir );
	foreach $file (readdir( DIR )) {
	    next unless( (-f $dir."/".$file) && ($file =~ /\./) );
	    print "reading state file $file\n" if( $main::dl>8 );
	    open( F, "<$dir/$file" );
	    $fulltext = $state = <F>;
	    close F;
	    chomp $fulltext;
	    $state =~ /^([a-z]*)/;
	    $state = $bb_evgen::numstates{$1};
	    print "state: $state\n" if( $main::dl>15 );
	    $prev = $bb_evgen::previous_state{$file};
	    if( $prev != $state ) {
		my( $host, $item ) = split( /\./, $file );
		($prev = $state) unless( defined $prev );
		&bb_evgen::raise( $host, $item, $prev, $state, $fulltext );
	    }
	    $bb_evgen::previous_state{$file} = $state;
	}
	close DIR;
    }
}
    

sub page {
    my( $host, $code, $text, $severity, $state, $fulltext, $msgmax, @users ) = @_;
    my( $short );

    $fulltext =~ s/^.*?[0-9]{4}\s//;
    $short = $text." ".$state.":".$fulltext;
    $short = substr( $short, 0, $msgmax ) if( $msgmax );
    $fulltext = $text." ".$state.":".substr($fulltext,0,256);
    ($severity = 50) unless( $severity ); 
    (@users = ($bb_evgen::page_user)) unless( @users );
    system( $bb_evgen::pager, "-S", $severity, "-s", $short, 
		"-h", $host, "-m", $fulltext,
		@users );
}



sub raise {
    my( $host, $item, $old_state, $new_state, $text ) = @_;
    my( $addr, @matches, %vars, $match, $ref, $var, $cont, $up );

    print "raise( $host, $item, $old_state, $new_state, $text )\n" if( $main::dl>2 );
    if( $old_state<$new_state ) {
	my( $i );

	for( $i=$#bb_evgen::new_alarms; $i>=0; $i-- ) {
	    $var = $bb_evgen::new_alarms[$i];
	    if( (($$var{"host"} eq $host) && ($$var{"item"} eq $item)
		  && ($$var{"down"}<$new_state))
		|| ! eval( $$var{"check"} ) ) {
	        splice( @bb_evgen::new_alarms, $i, 1 );
	    }
	}
	for( $i=$#bb_evgen::alarms; $i>=0; $i-- ) {
	    $var = $bb_evgen::alarms[$i];
	    if( (($$var{"host"} eq $host) && ($$var{"item"} eq $item)
		  && ($$var{"down"}<$new_state)) ) {
	        &bb_evgen::clear_alarm( $i );
	    }
	}
    }
    my ($name,$aliases,$addrtype,$length,@addrs) = gethostbyname( $host );
    foreach $addr (@addrs) {
	$addr = inet_ntoa( $addr );
    }
    @matches = grep( &match_rule( $_, $host, $item, \@addrs ), @bb_evgen::rules );
    %vars = ( "host", $host, "item", $item, "text", $text );
    foreach $match (@matches) {
	 $ref = $$match[2];
	 while( ($var,$cont) = each %$ref ) {
	     $vars{$var} = $cont;
	 }
    }
    if( $main::dl>3 ) {
	print "matched rule: ";
	while( ($var,$cont) = each %vars ) {
	    print "$var=$cont, ";
	}
	print "\n";
    }
    unless( defined $vars{"down"} ) {
	    $vars{"down"} = 1;
    }
    $bb_evgen::up{$host.".".$item} = $new_state>$vars{"down"};
    return if( ($old_state <= $new_state) || $up );
    if( $vars{"check"} ) {
	$vars{"check"} = &bb_evgen::compile( $vars{"check"}, %vars );
    }
    else {
	$vars{"check"} = 1;
    }
    return unless( eval $vars{"check"} );
    if( grep( ($host eq $$_{"host"}) && ($item eq $$_{"item"}), 
	@bb_evgen::new_alarms, @bb_evgen::alarms, @bb_evgen::old_alarms ) ) {
	print "no alarm since alarm pending/norepeated\n" if( $main::dl>2 );
	return;
    }
    if( $vars{"keep"} ) {
	$vars{"keep"} = time + $bb_evgen::minute*$vars{"keep"};
    }
    unless( $vars{"repeatprio"} ) {
	$vars{"repeatprio"} = $vars{"prio"};
    }
    if( $vars{"delay"} ) {
	$vars{"delay"} = time+$bb_evgen::minute*$vars{"delay"};
	push( @bb_evgen::new_alarms, \%vars );
	print "goes to delayed alarm queue\n" if( $main::dl>2 );
    }
    else {
	&bb_evgen::alarm( \%vars );
    }
}



sub alarm {
    my( $vars, $repeat ) = @_;

    print "\x07raising alarm: ".$$vars{"host"}.".".$$vars{"item"}." ".$$vars{"text"}."\n" if( $main::dl );
    push( @bb_evgen::alarms, $vars ) unless( grep( $_ eq $vars, @bb_evgen::alarms ) );
    if( $$vars{"repeat"} ) {
	($$vars{"_next_repetition"} = time) unless( $$vars{"_next_repetition"} );
	$$vars{"_next_repetition"} += $$vars{"repeat"}*$bb_evgen::minute;
    }
    if( $$vars{"norepeat"} ) {
	$$vars{"norepeat"} = time + $bb_evgen::minute*$$vars{"norepeat"};
    }
    &bb_evgen::page( $$vars{"host"}, $$vars{"item"}, $$vars{"host"}.":".$$vars{"item"}, $repeat?$$vars{"repeatprio"}:$$vars{"prio"}, $repeat?"reminder":"down", $$vars{"text"}, $$vars{"msgmax"}, split(",",$$vars{"mail"}) );
}



sub clear_alarm {
    my( $i ) = @_;
    my( $vars );

    $vars = $bb_evgen::alarms[$i];
    print "\x07clearing alarm: ".$$vars{"host"}.".".$$vars{"item"}." ".$$vars{"text"}."\n" if( $main::dl );
    if( $$vars{"keep"} && ($$vars{"keep"}>time) ) {
	$$vars{"_dequeue"} = $$vars{"keep"};
	print "keep alarm since 'keep' was defined\n" if( $main::dl>1 );
	return;
    }
    splice( @bb_evgen::alarms, $i, 1 );
    if( $$vars{"norepeat"} && ($$vars{"norepeat"}>time) ) {
	push( @bb_evgen::old_alarms, $vars );
    }
    &bb_evgen::page( $$vars{"host"}, $$vars{"item"}, $$vars{"host"}.":".$$vars{"item"}, $$vars{"prio"}, "up", "", $$vars{"msgmax"}, split(",",$$vars{"mail"}) );
}


sub compile {
    my( $check, %vars ) = @_;
    my( $s );

    $s = "";
    print "compile( $check )\n" if( $main::dl>3 );
    while( $check =~ /\$([a-z]+)/ ) {
	$s .= $`.$vars{$1};
	$check = $';
    }
    $check = $s.$check;
    $s = "";
    while( $check =~ /([a-z:_0-9\-]+)\.([a-z:_0-9\-]+)/ ) {
	$s .= $`."(\$bb_evgen::up{\"$1.$2\"})";
	$check = $';
    }
    $check = $s.$check;
    print "compiled to $check\n" if( $main::dl>4 );
    return $check;
}



sub match_rule {
    my( $rule, $host, $item, $addrs ) = @_;
    my( $rhost, $ritem, $vars ) = @$rule;
    my( @bla );

    return(0) if( ($ritem ne "*") && ($ritem ne $item) );
    if( substr( $rhost, 0, 1 ) eq "[" ) {
	$rhost = substr( $rhost, 1, length($rhost)-2 );
	return(0) unless( grep( substr( $_.".", 0, length($rhost)+1 ) eq $rhost.".", @$addrs ) );
    }
    else {
	return(0) if( ($rhost ne "*") && ($rhost ne $host) );
    }
    return 1;
}




sub load_rules {
    my( $var, $match, $fh, $chk );
    my( $host, $item );

    $bb_evgen::rule_file_stat = (stat($bb_evgen::rule_file))[9];
    unless( -f "$bb_evgen::rule_file" ) {
	print STDERR "cannot load rule file $bb_evgen::rule_file\n";
	return 0;
    }
    print "loading rule file $bb_evgen::rule_file\n" if( $main::dl>1 );
    @bb_evgen::rules = ();
    $fh = FileHandle->new();
    open( $fh, "<$bb_evgen::rule_file" );
    while( <$fh> ) {
	chomp;
	next if( /^#/ || /^$/ );
	while( /\\$/ ) {
	    $chk = $_;
	    $_ = <$fh>;
	    last unless( $_ );
	    chomp;
	    $_ = substr($chk,0,length($chk)-1).$_;
	}
	if( /^([^\s\t]+)/ ) {
	    $match = $1;
	    $var = $';
	}
	else {
	    next;
	}
	my %vars = ();
	while( $var ) {
	    $var =~ s/^[\s\t]+//;
	    if( ($var =~ /^([a-z0-9]+)[\s\t]*=[\s\t]*\"([^\"]+)\"/)
	    	|| ($var =~ /^([a-z0-9]+)[\s\t]*=[\s\t]*([^\s\t]+)/) ) {
		$vars{$1} = $2;
		$var = $';
	    }
	    else {
		last;
	    }
	}
	if( $vars{"down"} ) {
	    $vars{"down"} = $bb_evgen::numstates{$vars{"down"}};
	}
	foreach $chk (split( ";", $match )) {
	    $chk =~ /^(.*)\.([^\.]*)$/;
	    ($host, $item) = ($1, $2);
	    push( @bb_evgen::rules, [ $host, $item, \%vars ] );
	}
    }
    close $fh;
    return 1;
}


sub open_logs {
    my @log_files = @_;
    my $log_file;
    my( $open, $ino );

    @bb_evgen::logs = ();
    foreach $log_file (@log_files) {
	 my( $fh, $pos ) = (FileHandle->new(), 0);
	 my $open = 0;
	 if( open( $fh, "<$log_file" ) ) {
	     $open = 1;
	     seek( $fh, 0, 2 );
	     $pos = tell($fh);
	     $ino = ((stat($log_file))[1]);
	}
	push( @bb_evgen::logs, [ $log_file, $fh, $pos, $open, $ino, "" ] );
    }
}

1;
