#!/usr/bin/perl -I/usr/local/lib/bs/bin
#    Big Sister network monitor
#    Copyright (C) 1998  Thomas Aeby
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

#=============================================================================
#
my $Usage	  = "[-D level] [-c configfile] [-b bigsister-root]";
#
#=============================================================================

# Trim the program path name
$prog = $progdir = $0;
$prog =~ s#.*/##;
$progdir =~ s#/[^\/]*$##;
$progdir =~ /^(.*)$/; $progdir = $1; # untaint ...
 
use Getopt::Long;
use strict;
use FileHandle;
use GD;
use POSIX;

my %readers = ();
my %groups = ();

$Getopt::Long::ignorecase = 0;  
$main::opt_D = 0;
GetOptions( "D:i", "b:s", "c:s", "d:i" ) || &usage;
my $dl = $main'dl = $main::opt_D;
my $config = $main::opt_c;
my $goback = $main::opt_d;
$config = "reporting.cfg" unless( $config );

my %stat2text = ( 0, "red", 1, "yellow", 2, "purple", 3, "green" );

my $group_components = {};
my $done_statistics = {};

my $timezone = (12-((localtime(12*3600))[2]))*3600;
my $range_morning = 7*3600;
my $range_evening = 17*3600;

if( $main::opt_b ) {
    # untaint ...
    $main::opt_b =~ /^(.*)$/;
    $main::root = $1;
}
elsif( -d "/l/lib/bs" ) {
    $main::root = "/l/lib/bs";
}
elsif( -d "/usr/local/lib/bs" ) {
    $main::root = "/usr/local/lib/bs";
}
else {
    $main::root = $main::progdir."/..";
}

unshift( @INC, "$main::root/bin" );


eval "use Stat::stat";
my $status = new Stat::stat;

error( "unable to read config file $config: $!" ) unless( open( CFG, "<$config" ) );

while( <CFG> ) {
    chomp;
    next if( /^$/ || /^[\s\t]*#/ );
    if( /^read[\s\t]+([a-z\-_0-9]+)[\s\t]*/ ) {
	my @stats;
	my $reader = $1;
	my @args = split( /[\s\t]+/, $' );
	unless( $readers{$reader} ) {
	    eval "use Reader::$reader";
	    $readers{$reader} = 1;
	}
	$reader = eval "new Reader::$reader(".join(",",@args).")";
	error( "reader $reader won't start at line $. - bad arguments?" ) unless( $reader );
	while( @stats = $reader->get ) {
	    $status->put( @stats );
	}
	my $group;
	my $ref = $status->{"hist"};
	foreach $group (keys %$ref) {
	    $groups{$group} = {};
	}
    }
    elsif( /^prefix[\s\t]+(.*)$/ ) {
	$status->prefix( ($1 eq "none")?"":$1 );
    }
    elsif( /^group[\s\t]+(.*)/ ) {
	my $group = $1;
	my %components = ();
	$groups{$group} = {};
	print STDERR "creating group $group\n" if( $dl );
	while( <CFG> ) {
	    chomp;
	    next if( /^$/ || /^[\s\t]*#/ );
	    last unless( /^[\s\t]+/ );
	    if( /^[\s\t]+(add|remove)[\s\t]+(.*)/ ) {
		my $component;
		my $cmd = $1;
		my @components = &components( $2 );
		foreach $component (@components) {
		    print STDERR "    $cmd $component\n" if( $dl>4 );
		    $components{$component} = ($cmd eq "add");
		}
	    }
	    elsif( /^[\s\t]+set[\s\t]+(.*?)=(.*)$/ ) {
		$groups{$group}->{$1} = $2;
	    }
	    elsif( /^[\s\t]+init[\s\t]+(\d)/ ) {
		$status->put( [ $group, $1, 0, time, "defaults to $1" ] );
	    }
	    else {
		print STDERR "unknown group subcommand at line $.\n";
	    }
	}
	my $next_line = $_;
	my @stat = ();
	my $c;
	my $cref;

        $cref = $group_components->{$group} = [];
	foreach $c (keys %components) {
	    next unless $components{$c};
	    print STDERR "  adding sub $c\n" if( $dl>1 );
	    push( @$cref, $c );
	    my $stat = $status->{"hist"}->{$c};
	    next unless( $stat );
	    push( @stat, @$stat );
	}
	@stat = sort { $$b[1] <=> $$a[1] } @stat;
	foreach $c (@stat) {
	    my( $item, $stat, $from, $to, $comment ) = @$c;
	    $status->put( [ $group, $stat, $from, $to, $comment ] );
	}
	$status->optimize( $group );
	$_ = $next_line;
	redo;
    }
    elsif( /^statistics[\s\t]+(.*)$/ ) {
	my $group = $1;
	my $days = 30;
	my $steps = 30;
	my ( $x, $y );
	my ( $tx, $ty );
	my $tsteps = 90;
	my $minperc = 60;
	while( <CFG> ) {
	    chomp;
	    next if( /^$/ || /^[\s\t]*#/ );
	    last unless( /^[\s\t]+/ );
	    if( /^[\s\t]+days[\s\t]+(.*)/ ) {
		$days = $1;
	    }
	    elsif( /^[\s\t]+steps[\s\t]+(.*)/ ) {
		$steps = $1;
	    }
	    elsif( /^[\s\t]+graph[\s\t]+(.*)/ ) {
		($x, $y) = split( /[\sxX]/, $1 );
	    }
	    elsif( /^[\s\t]+minpercent[\s\t]+(.*)/ ) {
		$minperc = $1;
	    }
	    elsif( /^[\s\t]+timegraph[\s\t]+(.*)/ ) {
		($tx, $ty, $tsteps ) = split( /[\sxX]/, $1 );
		($tsteps = 90) unless( $tsteps );
	    }
	    else {
		print STDERR "unknown group subcommand at line $.\n";
	    }
	}
        statistics( $group, $days, $steps, $x, $y, $tx, $ty, $tsteps, $minperc );
    }
}

my %html_done = ();
my $group;
foreach $group (keys %$done_statistics) {
    $group =~ s/ .*//;
    next if( $html_done{$group} );
    $html_done{$group} = 1;
    &do_html( $group );
}


sub do_html {
    my( $group ) = @_;
    my( $key );
    my %vars;

    %vars = ( "GROUP", $group, "STATLIST", "", "AVAIL", "" );
    foreach $key ( sort keys %$done_statistics ) {
	next unless( $key =~ /\Q$group / );
	$key =~ s/$group //;
	if( $key =~ /^(\d+)\s+(\d+)/ ) {
	    my( $days, $period ) = ($1, $2);
	    $vars{"AVAIL"} .= "<P>\n<H3>$days days availability</H3>\n<IMG SRC=\"$days-$period.gif\">\n</P>\n";
	}
	elsif( $key =~ /^stat (\d+)/ ) {
	    my( $n ) = $1;
	    $vars{"STATLIST"} .= " <P> \n<IMG SRC=\"statlist$n.gif\"> \n</P> \n";
	}
    }
    $vars{"COMPONENTS"} = "";
    my $components = $group_components->{$group};
    if( $components ) {
	foreach $key (sort @$components) {
	    $vars{"COMPONENTS"} .= "<A HREF=\"../$key/index.html\">$key</A><BR>\n";
	}
    }
    my $text = `cat template.proto`;
    foreach $key (keys %vars) {
	$text =~ s/\@$key\@/$vars{$key}/g;
    }
    open( OUT, ">data/$group/index.html" );
    print OUT $text;
    close OUT;
}


sub components {
    my( $str ) = @_;
    my( $c, @cs, $i );

    print STDERR "      components( $str )\n" if( $dl>5 );
    @cs = split( /[\s\t]+/, $str );
    for( $i=0; $i<=$#cs; $i++ ) {
	$c = $cs[$i];
	while( $c =~ /\{([^\}]*)\}/ ) {
	    my ($pre,$cd,$post) = ($`,$1, $');
	    my @sub = split( ",", $cd );
	    my $sub = shift @sub;
	    $c = $pre.$sub.$post;
	    foreach $sub (@sub) {
		push( @cs, $pre.$sub.$post );
	    }
	}
	if( $c =~ /[\?\*]/ ) {
	    $c =~ s/([\.\(\)\{\}])/\\$1/g;
	    $c =~ s/\*/\.\*/g;
	    $c =~ s/\?/\./g;
	    my @match = grep( /^$c$/, keys %groups );
	    splice( @cs, $i, 1, @match );
	    $i += $#match;
	}
	else {
	    $cs[$i] = $c;
	}
    }
    return @cs;
}



sub error {
    my( $error ) = @_;

    print STDERR $error."\n";
    exit(1);
}


sub statistics {
    my( $groups, $days, $period, $x, $y, $tx, $ty, $tdays, $minperc ) = @_;
    my @groups = split( ",", $groups );
    my( $base, $dc, $di, $stat, %sum, $partial );
    my( $i, $image );
    my( $tstart, @stats, $y1prev, $y2prev );

    for( $i=0; $i<=$#groups; $i++ ) {
        my $group = $groups[$i];
        next if( $done_statistics->{"$group $days $period"} );
        $done_statistics->{"$group $days $period"} = 1;
        $stat = $status->{"hist"}->{$group};
        next unless $stat;
	print "computing statistics for $group\n" if( $dl );

        my $cref = $group_components->{$group};
        if( $cref ) {
            my $i;
            foreach $i (@$cref) {
                push( @groups, $i );
            }
        }
        my $day_start = time-$timezone-$goback*24*3600;
        $day_start -= $day_start%(24*3600);
        $base = "data/$group/$days-$period";
	mkdir( "data/$group", 0755 );
        open( OUT, ">$base.csv" );

	if( $x ) {
	    $image = &image_setup( $x, $y );
	    &prepare_time_axis( $image );
	    &prepare_percent_axis( $image );
	    &time_axis( $image, $day_start - $period*24*3600, $day_start );
	    &percent_axis( $image, $minperc, 100 );
	}

	undef $y1prev, $y2prev;
        for( $dc=$period; $dc; $dc-- ) {
	    my( $thisday, $thiswday );
            %sum = ("total", 0);
            for( $di=0; $di<$days; $di++ ) {
                my( $start, $end, $key );
                $thisday = $day_start - ($dc+$di)*24*3600;
                $start = $thisday + $range_morning;
		$thiswday = ((localtime($start))[6]);
		next if( ($thiswday > 5) || ($thiswday < 1) );
                $end = $thisday + $range_evening;
                $partial = check_interval( $start, $end, $stat );
                $sum{"total"} += $end-$start;
                foreach $key (keys %$partial ) {
                    $sum{$key} = 0 unless( defined $sum{$key} );
                    $sum{$key} += $partial->{$key};
                }
            }
	    $thisday = $day_start - $dc*24*3600;
            print OUT join(";", $thisday,
                $sum{0}, $sum{1}, $sum{2},
		$sum{3}+$sum{1}, int( ($sum{3}+$sum{1})*1000/$sum{"total"} + 0.5 )/10,
		$sum{2}+$sum{3}+$sum{1}, int( ($sum{2}+$sum{3}+$sum{1})*1000/$sum{"total"} + 0.5)/10,
		$sum{"total"}
            )."\n";

	    if( $x ) {
		my $poly = new GD::Polygon;
		my $x = &xpos( $image, $thisday-24*3600 );
		my $xn = &xpos( $image, $thisday );
		my $y1 = &ypos( $image, ($sum{2}+$sum{3}+$sum{1})*100/$sum{"total"} );
		my $y2 = &ypos( $image, ($sum{3}+$sum{1})*100/$sum{"total"} );
		if( defined $y1prev ) {
		    $y2++ if( $y1 == $y2 );
		    $poly->addPt( $x, $y1prev );
		    $poly->addPt( $xn, $y1 );
		    $poly->addPt( $xn, $y2 );
		    $poly->addPt( $x, $y2prev );
		    $image->{"image"}->filledPolygon( $poly, $image->{"blue"} );
		}
		$y1prev = $y1;
		$y2prev = $y2;
	    }
        }
        close OUT;

	if( $x ) {
	    &time_axis( $image, $day_start - $period*24*3600, $day_start );
	    &percent_axis( $image, $minperc, 100 );
	    open( OUT, ">$base.gif" );
	    print OUT $image->{"image"}->gif;
	    close OUT;
	}

	open( OUT, ">data/$group/statlist.csv" );
	my( $i );
	foreach $i (@$stat) {
	    my( $group, $statc, $from, $to, $comment ) = @$i;
	    $comment =~ tr/;/,/;
	    print OUT join( ";",
		$from, scalar localtime($from), 
		$to, scalar localtime($to),
		$stat2text{$statc}, $comment )."\n";
	}
	close OUT;

	if( $tx ) {
	    my $n;
	    my $day_start = $day_start;

	    for( $n=0; $n<4; $n++ ) {
		next if( $done_statistics->{"$group stat $n"} );
		$done_statistics->{"$group stat $n"} = 1;
		$image = &image_setup( $tx, $ty );
		&prepare_time_axis( $image );
		&time_axis( $image, $tstart = $day_start - $tdays*24*3600, $day_start );
		@stats = ();

		@stats = sort { $$b[1] <=> $$a[1] } @$stat;
		foreach $i (@stats) {
		    my( $group, $statc, $from, $to, $comment ) = @$i;
		    $from = $tstart if( $from<$tstart );
		    $to = $day_start if( $to>$day_start );
		    next if( $from>$to );
		    $from = &xpos( $image, $from );
		    $to = &xpos( $image, $to );
		    $image->{"image"}->filledRectangle( $from, $image->{"y0"}, $to, $image->{"ymax"}, $image->{$stat2text{$statc}} );
		}
		&time_axis( $image, $day_start - $tdays*24*3600, $day_start );
		open( OUT, ">data/$group/statlist$n.gif" );
		print OUT $image->{"image"}->gif;
		close OUT;
		$day_start = $tstart;
	    }
	}
    }
}



sub image_setup {
    my( $x, $y ) = @_;
    my( $im );
    my $image = {};

    $im = $image->{"image"} = new GD::Image( $x, $y );
    $image->{"white"} = $im->colorAllocate(255,255,255);
    $image->{"meta"} = $image->{"black"} = $im->colorAllocate(0,0,0);
    $image->{"green"} = $im->colorAllocate(0,255,0);
    $image->{"blue"} = $im->colorAllocate(0,0,255);
    $image->{"red"} = $im->colorAllocate(255,0,0);
    $image->{"yellow"} = $im->colorAllocate(255,255,0);
    $image->{"purple"} = $im->colorAllocate(128,0,255);
    $image->{"font"} = gdSmallFont;
    $image->{"fontwidth"} = $image->{"font"}->width;
    $image->{"fontheight"} = $image->{"font"}->height;
    $im->filledRectangle( 0,0,$x,$y,$image->{"white"} );
    $image->{"x0"} = 10;
    $image->{"y0"} = 10;
    $image->{"xmax"} = $x-10;
    $image->{"ymax"} = $y-10;
    return $image;
}


sub prepare_percent_axis {
    my( $image ) = @_;
    $image->{"x0"} += $image->{"fontwidth"}*4 + 10;
    $image->{"y0"} += $image->{"fontheight"}/2;
}
   

sub prepare_time_axis {
    my( $image ) = @_;
    $image->{"ymax"} -= $image->{"fontheight"} + 10;
    $image->{"xmax"} -= $image->{"fontwidth"}*3;
}
   

sub percent_axis {
    my( $image, $start, $end ) = @_;
    my $im = $image->{"image"};
    my( $per, $len );

    $im->line( $image->{"x0"}, $image->{"ymax"}, $image->{"x0"}, $image->{"y0"}, $image->{"meta"} );
    $image->{"ystart"} = $start;
    $image->{"ystarty"} = $image->{"ymax"};
    $image->{"yend"} = $end;
    $image->{"yendy"} = $image->{"y0"};
    $per = int(($start+9)/10)*10;
    for( $per = $start; $per<=$end; $per+=1 ) {
	my $ypos = &ypos( $image, $per );
	if( $per %10 == 0 ) {
	    $im->line( $image->{"x0"}, $ypos, $image->{"x0"}-10, $ypos, $image->{"meta"} );
	    my $str = $per."%";
	    $im->string( $image->{"font"}, $image->{"x0"} - 11 - ($image->{"fontwidth"}*length($str)),
			 $ypos - $image->{"fontheight"}/2, $str, $image->{"meta"} );
	}
	elsif( $per % 5 == 0 ) {
	    $im->line( $image->{"x0"}, $ypos, $image->{"x0"}-9, $ypos, $image->{"meta"} );
	}
	else {
	    $im->line( $image->{"x0"}, $ypos, $image->{"x0"}-7, $ypos, $image->{"meta"} );
	}
    }
}


sub time_axis {
    my( $image, $start, $end ) = @_;
    my $im = $image->{"image"};
    my( $day, $len );

    $im->line( $image->{"x0"}, $image->{"ymax"}, $image->{"xmax"}, $image->{"ymax"}, $image->{"meta"} );
    $image->{"xstart"} = $start;
    $image->{"xstartx"} = $image->{"x0"};
    $image->{"xend"} = $end;
    $image->{"xendx"} = $image->{"xmax"};
    for( $day = $start; $day<=$end; $day+=24*3600 ) {
	my $xpos = &xpos( $image, $day );
	if( ((localtime($day))[6]) ) {
	    $im->line( $xpos, $image->{"ymax"}, $xpos, $image->{"ymax"}+7, $image->{"meta"} );
	}
	else {
	    $im->line( $xpos, $image->{"ymax"}, $xpos, $image->{"ymax"}+10, $image->{"meta"} );
	    my $str = strftime( "%h %d", localtime($day) );
	    $im->string( $image->{"font"}, $xpos - ($image->{"fontwidth"}*length($str)/2),
			 $image->{"ymax"}+11, $str, $image->{"meta"} );
	}
    }
}


sub ypos {
    my( $image, $val ) = @_;

    return( $image->{"ystarty"} + ($image->{"yendy"}-$image->{"ystarty"}) 
				* ($val-$image->{"ystart"}) 
				/ ($image->{"yend"}-$image->{"ystart"}) );
}


sub xpos {
    my( $image, $val ) = @_;

    return( $image->{"xstartx"} + ($image->{"xendx"}-$image->{"xstartx"}) 
				* ($val-$image->{"xstart"}) 
				/ ($image->{"xend"}-$image->{"xstart"}) );
}



sub check_interval {
    my( $start, $end, $stat ) = @_;
    my( %partial, $ref );
    my( $group, $statc, $from, $to, $comment );

    foreach $ref (@$stat) {
        ( $group, $statc, $from, $to, $comment ) = @$ref;
        $from=$start if( $from<$start );
        $to=$end if( $end<$to );
        if( $to>$from ) {
            $partial{$statc} = 0 unless( defined $partial{$statc} );
            $partial{$statc} += $to-$from;
        }
    }
    return \%partial;
}

